import os, random, pygame, math
from Numeric import *
from pygame.locals import *
#from filters import ColorMatrixFilter
from transitions import Tween
from transitions.easing import Linear
import pygame.surfarray as surfarray

if not pygame.font: print 'Warning, fonts disabled'
if not pygame.mixer: print 'Warning, sound disabled'

#
#Tile.py
#############################

class Tile(pygame.sprite.Sprite):
	tileType = "rock"
	tileImage = None
	tileRect = None
	movable = True
	screenPosition = (0, 0)

	def __init__(self, t, img):
		pygame.sprite.Sprite.__init__(self)
		self.tileType = t
		self.tileImage = img
		self.tileRect = img.get_rect()

	def clone(self):
		return Tile(self.tileType, self.tileImage)

class Selector(pygame.sprite.Sprite):
	def __init__(self):
		#cmf = ColorMatrixFilter(array(([1,0,1,0],[0,1,0,0],[1,0,1,0],[50,50,50,1])))
		self.image = load_image(os.path.join('data', 'Selector.png') )
		self.rect = self.image.get_rect()
		self.position = (0, 0)
		self.tile = None
		
		#cmf.applyFilter(surfarray.pixels3d(self.image))
		

#
#World.py
#############################
class World:
	orientation = "north"
	depth = 0
	width = 0
	height = 0
	windowSize = None
	offset = (0, 0)
	
	tileWidth = 100
	tileDepth = 80
	tileHeight = 40
	tileImageHeight = 171
	
	shadowImages = None
	shadows = None # Has the same dimensions as terrain. Each position contains
	               # a list with the shadows to be applied to a tile space.
	
	terrain = [] # One-dimensional array, length == depth * width.
				 # Width varies faster.
				 # Each position has an array representing height.
	
	selector = None
	
	def __init__(self, depth, width, height, shadowImages, terrain = []):
		self.depth = depth
		self.width = width
		self.height = height
		self.selector = Selector()
		self.shadowImages = shadowImages
		
		if terrain == []:
			for i in range(depth*width):
				self.terrain.append([])
		else:
			for i in range(depth*width):
				self.terrain.append(terrain[i])
		
		self.windowSize = (self.tileWidth*width, self.tileHeight*height + self.tileDepth*(depth+1))
		self.update(self.offset)
	
	def getPosition(self, (x, y)):
		return y*self.width + x
	
	def update(self, offset):
		self.offset = offset
		for i in range(self.depth):
			for j in range(self.width):
				tileSpace = self.terrain[i*(self.width)+j]
				for k in range(len(tileSpace)):
					tile = tileSpace[k]
					x = offset[0] + j*self.tileWidth
					y = offset[1] + i*self.tileDepth+(self.height - k)*self.tileHeight
					tile.screenPosition = (x, y);
					tile.tileRect.topleft = (x,y+self.tileImageHeight-self.tileDepth-self.tileHeight)
					tile.tileRect.size = (self.tileWidth, self.tileHeight+self.tileDepth)
		self.updateShadows()
	
	def handAction(self):
		tileSpace = self.terrain[self.getPosition(self.selector.position)]
		if self.selector.tile:
			if len(tileSpace) < self.height:
				tileSpace.append(self.selector.tile)
				self.selector.tile = None
		else:
			if tileSpace[-1].movable:
				self.selector.tile = tileSpace[-1]
				self.terrain[self.getPosition(self.selector.position)] = tileSpace[0:len(tileSpace)-1]
		self.update(self.offset)
	
	# setSelector, getMouseIntersect and getIntersectRectangle are the 3
	# functions used to implement the selector cursor
	
	def setSelector(self, (x, y)):
		"""Given (x,y) as a coordinate on the terrain, puts the selector cursor
		   on its corresponding space, and updates its position for the next
		   draw pass on screen"""
		self.selector.position = (max(0, min(x, self.width-1)), max(0, min(y, self.height-1)))
		selectorTileSpace = self.terrain[self.selector.position[1]*(self.width)+self.selector.position[0]]
		
		self.selector.rect.topleft = selectorTileSpace[-1].screenPosition
		self.selector.rect.top = self.selector.rect.top - self.tileHeight
	
	def getMouseIntersect(self, (x, y)):
		"""Given (x,y) as a mouse position relative to the terrain (i.e. top left
		   corner of the top left tile is (0, 0)) returns the map coordinates of
		   the tile under that position. If the cursor is outside the map it will
		   try to approximate as best as it can.
		   
		   If x is inside the map width, the cursor will follow the mouse.
		   Otherwise, it will hang at the border of the map waiting for the
		   mouse to reenter."""
		# FIXME: Will have to change when adjust to different orientations
		columnPosition = max(0, min(int( math.floor( x/self.tileWidth ) ), self.width-1))
		rowIntersectionBoxes = [self.getIntersectRectangle(self.getPosition((columnPosition, tileY))) for tileY in range(self.depth)]
		rowIntersectsMouse = map(lambda r: r.collidepoint((x, y)), rowIntersectionBoxes)
		
		if any(rowIntersectsMouse):
			for i in reversed(range(self.depth)):
				if (rowIntersectsMouse[i]):
					return (columnPosition, i)
		else:
			bigIntersectionBox = reduce(lambda ra, rb: ra.union(rb), rowIntersectionBoxes)
			if (y < bigIntersectionBox.top):
				return (columnPosition, 0)
			elif (y > bigIntersectionBox.bottom):
				return (columnPosition, self.depth-1)
			else:
				return self.selector.position
		
	
	def getIntersectRectangle(self, listPosition):
		"""Given a tile space, it will get the hit rectangles from each tile
		   and will return the union of all them."""
		if (len(self.terrain[listPosition]) == 0):
			return Rect(0,0,0,0)
		return self.terrain[listPosition][0].tileRect.unionall( map( lambda x: x.tileRect, self.terrain[listPosition][1:] ) )
	
	# updateShadows, getTileHeight, getCorner and detectShadows are the 4
	# functions used for the shadowing algorithm
	
	def updateShadows(self):
		"""For the current terrain configuration detect the shadows inciding in
		   each position. Only to be called when modifying terrain or rotating it"""
		# FIXME: Will have to change this to adjust to different orientations
		terrainIndices = [(x, y) for y in range(self.depth) for x in range(self.width)]
		
		nw = map(self.getTileHeight, map(lambda x: self.getCorner(x, (-1, -1)), terrainIndices))
		n  = map(self.getTileHeight, map(lambda x: self.getCorner(x, ( 0, -1)), terrainIndices))
		ne = map(self.getTileHeight, map(lambda x: self.getCorner(x, ( 1, -1)), terrainIndices))
		w  = map(self.getTileHeight, map(lambda x: self.getCorner(x, (-1,  0)), terrainIndices))
		e  = map(self.getTileHeight, map(lambda x: self.getCorner(x, ( 1,  0)), terrainIndices))
		sw = map(self.getTileHeight, map(lambda x: self.getCorner(x, (-1,  1)), terrainIndices))
		s  = map(self.getTileHeight, map(lambda x: self.getCorner(x, ( 0,  1)), terrainIndices))
		se = map(self.getTileHeight, map(lambda x: self.getCorner(x, ( 1,  1)), terrainIndices))
		self.shadows = map(self.detectShadows, map(self.getTileHeight, terrainIndices), nw, n, ne, w, e, sw, s, se, terrainIndices)

	def getTileHeight(self, p):
		"""Returns the height of a tile space. Returns 0 if p is None"""
		if p == None:
			return 0
		return len(self.terrain[p[1]*self.width+p[0]])
	
	def getCorner(self, pos, offset):
		"""For a ordered pair pos return its position shifted by offset. If
		   the position falls outside of the terrain returns None"""
		if pos[0] + offset[0] < 0 or pos[1] + offset[1] < 0 or \
		   pos[0] + offset[0] >= self.width or pos[1] + offset[1] >= self.depth:
			return None
		else:
			return (pos[0] + offset[0], pos[1] + offset[1])
	
	def detectShadows(self, p, nw, n, ne, w, e, sw, s, se, ind):
		"""returns for each p the list of shadows it has to apply to p"""
		shadowSquare = dict()
		
		shadowSquare['nw'] = nw > p and w <= p and n <= p
		shadowSquare['n'] = n > p
		shadowSquare['ne'] = ne > p and e <= p and n <= p
		shadowSquare['w'] = w > p
		shadowSquare['e'] = e > p
		shadowSquare['sw'] = sw > p and w <= p
		shadowSquare['s'] = s > p
		shadowSquare['se'] = se > p and e <= p
		shadowSquare['sidew'] = sw >= p
		shadowSquare['topmost'] = p == self.height
		
		return [self.shadowImages[key] for key in shadowSquare.keys() if shadowSquare[key]]

	def draw(self, screen):
		#This is for drawing for "north" orientation only
		for i in range(self.depth):
			for j in range(self.width):
				tileSpace = self.terrain[i*(self.width)+j]
				shadow = self.shadows[i*(self.width)+j]
				for k in range(len(tileSpace)):
					tile = tileSpace[k]
					
					screen.blit(tile.tileImage, tile.screenPosition)
					if k+1 == len(tileSpace):
						for l in shadow:
							screen.blit(l, tile.screenPosition)
						if (j, i) == self.selector.position:
							screen.blit(self.selector.image, self.selector.rect.topleft)
							if self.selector.tile:
								screen.blit(self.selector.tile.tileImage, (self.selector.rect.left, self.selector.rect.top-self.tileImageHeight))

def generateRandomTerrain(d, w, h, tiles):
	terr = []

	for i in range(d*w):
		tileSpace = []
		for j in range(0, random.randint(1, h)):
			tileSpace.append(random.choice(tiles).clone())
		terr.append(tileSpace)
	return terr

def generatePlainTerrain(d, w, h, tile, movable = True):
	terr = []

	for i in range(d*w):
		tileSpace = []
		for j in range(h):
			tileSpace.append(tile.clone())
			tileSpace[-1].movable = movable
		terr.append(tileSpace)
	return terr

#
#WorldScrollView.py
#############################
class WorldScrollView:
	view = None
	position = (0, 0)
	_viewPosition = (0, 0)
	__windowSize = None
	world = None
	border = (20, 30)
	
	def __init__(self, windowSize, world):
		self.view = pygame.Surface(windowSize, pygame.SRCALPHA, 32);
		self.__windowSize = windowSize
		self._viewPosition = (0, 0)
		self.world = world
		self.update()
	
	def update(self):
		self.view.fill((0, 0, 0, 0))
		self.world.draw(self.view)
	
	def draw(self, screen):
		screen.blit(self.view, self.position)
	
	def getViewPosition(self):
		return self._viewPosition*-1
	
	def setViewPosition(self, pos):
		self._viewPosition = (-max(0, min(self.world.windowSize[0]+self.border[0]*2-self.__windowSize[0], pos[0])),
		                      -max(0, min(self.world.windowSize[1]+self.border[1]*2-self.__windowSize[1], pos[1])))
		self.world.update(self._viewPosition)
	
	def setSelectorPosition(self, (mouseX, mouseY)):
		mapMouse = (mouseX-self.position[0], mouseY-self.position[1])
		selectorPos = self.world.getMouseIntersect(mapMouse)
		self.world.setSelector(selectorPos)

#
# CuteGod.py
###################################

def load_image(name, colorkey=None):
	fullname = name
	try:
		image = pygame.image.load(fullname)
	except pygame.error, message:
		print 'Cannot load image:', fullname
		raise SystemExit, message
	image = image.convert_alpha()
	if colorkey is not None:
		if colorkey is -1:
			colorkey = image.get_at((0,0))
		image.set_colorkey(colorkey, RLEACCEL)
	return image

def initShadowImages():
	shadowImages = dict()
	shadowImages['nw'] = load_image(os.path.join('data', 'Shadow North West.png'))
	shadowImages['n'] = load_image(os.path.join('data', 'Shadow North.png'))
	shadowImages['ne'] = load_image(os.path.join('data', 'Shadow North East.png'))
	shadowImages['w'] = load_image(os.path.join('data', 'Shadow West.png'))
	shadowImages['e'] = load_image(os.path.join('data', 'Shadow East.png'))
	shadowImages['sw'] = load_image(os.path.join('data', 'Shadow South West.png'))
	shadowImages['s'] = load_image(os.path.join('data', 'Shadow South.png'))
	shadowImages['se'] = load_image(os.path.join('data', 'Shadow South East.png'))
	shadowImages['sidew'] = load_image(os.path.join('data', 'Shadow Side West.png'))
	shadowImages['topmost'] = shadowImages['sidew']
	return shadowImages

class TweenParameters():
	def __init__(self, vOffset, pos):
		self.vOffset = vOffset
		self.pos = pos
	
	def getPosition(self):
		return (self.pos[0], self.pos[1]+self.vOffset)

def main():
	pygame.init()
	screen = pygame.display.set_mode( (800, 600) )
	clock = pygame.time.Clock()
	shadowImages = initShadowImages()

	backgroundImage = load_image(os.path.join('data', 'background.png'))
	selectorTile = load_image(os.path.join('data', 'Selector.png'))
	
	plainTile = Tile('plain', load_image(os.path.join('data', 'Plain Block.png') ))
	dirtTile = Tile('dirt', load_image(os.path.join('data', 'Dirt Block.png') ))
	grassTile = Tile('grass', load_image(os.path.join('data' , 'Grass Block.png') ))
	stoneTile = Tile('stone', load_image(os.path.join('data' , 'Stone Block.png') ))
	waterTile = Tile('water', load_image(os.path.join('data' , 'Water Block.png') ))

	terrain1 = generatePlainTerrain(6, 8, 1, plainTile, False)
	terrain2 = generateRandomTerrain(6, 8, 3, [dirtTile, grassTile, stoneTile, waterTile])
	terrain = map(lambda a, b: a + b, terrain1, terrain2)
	
	world = World(4, 6, 5, shadowImages, terrain)
	#world = World(3, 3, 5, shadowImages, [[stoneTile],[stoneTile,stoneTile],[stoneTile],[stoneTile],[stoneTile,stoneTile],[stoneTile],[stoneTile,stoneTile],[stoneTile],[stoneTile]])
	worldScrollView = WorldScrollView((600, 600), world)
	worldScrollView.position = (200, 0)
	worldScrollView.setViewPosition((0, 50))
	
	tp = TweenParameters(0, (0, 50))
	movementTween = Tween(tp, "vOffset", Linear.easeNone, 0, 50, 50, True, True)
	
	while 1:          
		screen.blit(backgroundImage, (0, 0))
		worldScrollView.draw(screen)
		worldScrollView.setViewPosition(tp.getPosition())
		worldScrollView.update()
		pygame.display.flip()

		for event in pygame.event.get():
			if event.type == QUIT:
				return
			elif event.type == KEYDOWN and event.key == K_ESCAPE:
				return
			elif event.type == MOUSEMOTION:
				worldScrollView.setSelectorPosition(event.pos)
			elif event.type == MOUSEBUTTONDOWN:
				#print "Click! Position: (%d, %d) Button %d" % (event.pos[0], event.pos[1], event.button)
				world.handAction()

		#print "View position: (%d, %d)" % tp.getPosition()
		movementTween.tick()
		clock.tick(60)
	
if __name__ == '__main__':
	main()

