class Tween:
	def __init__(self, object, property, function, begin, finish, duration, loop = False, yoyo = False):
		self.object = object
		self.property = property
		self.function = function
		self.begin = begin
		self.finish = finish
		self.position = begin
		self.duration = duration
		self.time = 0
		self.change = float(finish - begin)
		self.loop = loop
		self.yoyo = yoyo
		
		self.isPlaying = True
		self.started = False
		
		self.genFunc = self.functionGenerator()
	
	def functionGenerator(self):
		if self.loop:
			if self.yoyo:
				direction = 1
				while True:
					yield self.function(self.time, self.begin, self.change, self.duration)
					self.time += direction
					if self.time == self.duration or self.time == 0: direction *= -1
			else:
				while True:
					yield self.function(self.time, self.begin, self.change, self.duration)
					if self.time == self.duration: self.time = 0
					self.time += 1
		else:
			while self.time <= self.duration:
				yield self.function(self.time, self.begin, self.change, self.duration)
				self.time += 1
	
	def tick(self):
		if not self.isPlaying: return
		
		if not self.started:
			self.started = True
			if callable(self.onMotionStarted):
				self.onMotionStarted()
		
		try:
			self.position = self.genFunc.next()
		except StopIteration:
			self.isPlaying = False
			if callable(self.onMotionStopped):
				self.onMotionStopped()
			return
		
		setattr(self.object, self.property, self.position)
		if callable(self.onMotionChanged):
			self.onMotionChanged()
	
	onMotionStarted = None
	onMotionChanged = None
	onMotionStopped = None
